
from omnibelt import InitWall, unspecified_argument

class Configurable:
	'''
	Removes the config object `A` from the __init__() to clean up the init stream.
	
	This class should be subclassed when creating components/modifiers,
	especially when those components/modifiers also subclass types that
	do not use the config object.
	'''
	def __init__(self, A, _req_args=unspecified_argument,
	             _req_kwargs=unspecified_argument, **kwargs):

		if _req_args is unspecified_argument:
			_req_args = A.pull('_req_args', (), silent=True)
		if _req_kwargs is unspecified_argument:
			_req_kwargs = A.pull('_req_kwargs', {}, silent=True)

		# if _req_kwargs is None:
		# 	_req_kwargs = kwargs

		try:
			walled = isinstance(self, InitWall)
		except ValueError:
			walled = True
			
		if walled:
			super().__init__(_req_args=_req_args, _req_kwargs=_req_kwargs, **kwargs)
		else:
			kwargs.update(_req_kwargs)
			super().__init__(**kwargs)
			# super().__init__(*_req_args, **_req_kwargs)






import sys, os
from pathlib import Path
from collections import defaultdict, OrderedDict
from c3linearize import linearize
import importlib.util

from omnibelt import get_printer, load_yaml

from .containers import Container
from .rules import view_meta_rules, meta_rule_fns
from .errors import UnknownArtifactError, artifact_errors, MissingModifierError, AmbiguousRuleError, \
	MissingConfigError, MissingArtifactError
from .registry import Script_Registry, Component_Registry, Modifier_Registry, Config_Registry
from .external import include_files, include_package, register_project_type
from .util import global_settings, configurize, parse_arg
from . import projects as projects_module

prt = get_printer(__name__)


class Cerifiable: # TODO: turn this into a macro
	def __certify__(self, A, **kwargs):
		pass


class Workspace(Container):
	
	def __init__(self, silent=False, **kwargs):
		super().__init__(**kwargs)
		self.silent = silent
		
		self.reset_registries()
	
	def _process(self, raw):
		
		self.config_paths = raw.get('configs', [])
		if isinstance(self.config_paths, str):
			self.config_paths = [self.config_paths]
		self.src_paths = raw.get('src', [])
		if isinstance(self.src_paths, str):
			self.src_paths = [self.src_paths]
		self.src_packages = raw.get('packages', [])
		if isinstance(self.src_packages, str):
			self.src_packages = [self.src_packages]
		
		super()._process(raw)
	
	def initialize(self):
		'''
		This loads the project, primarily by registering any specified config files,
		importing specified packages, and finally running any provided source files

		:return: None
		'''
		
		self.load_configs(self.config_paths)
		self.load_src(self.src_paths, self.src_packages)
	
	def load_configs(self, paths=[]):
		'''Registers all specified config files and directories'''
		if len(paths):
			
			for path in paths:
				if os.path.isdir(path):
					self.register_config_dir(path, recursive=True)
				elif os.path.isfile(path):
					fname = os.path.basename(path)
					parts = fname.split('.')
					if len(parts) > 1 and parts[-1] in {'yml', 'yaml'}:
						self.register_config(parts[0], path)
	
	def load_src(self, srcs=[], packages=[]):
		'''Imports all specified packages and runs the specified python files'''
		include_package(*packages)
		include_files(*[src for src in srcs],)# project_name=self.get_name())

	def meta_rules(self):
		return view_meta_rules()

	def meta_rules_fns(self):
		return meta_rule_fns()
	
	# region Registration
	
	def reset_registries(self):
		'''Clears all registries'''
		if not self.silent:
			prt.debug(f'Resetting registries of {self}')
		self.scripts = Script_Registry()
		self.components = Component_Registry()
		self.modifiers = Modifier_Registry()
		self.configs = Config_Registry()
		
		self._registries = dict(script=self.scripts, component=self.components,
		                        modifier=self.modifiers, config=self.configs)
	
	def register_artifact(self, atype, name, info):
		'''General function to register an artifact of type `atype` with name `name`'''
		
		registry = self._registries.get(atype, None)
		if registry is None:
			raise UnknownArtifactError(atype)
		
		if not self.silent:
			msg_name = name if atype is None else f'{atype} {name}'
			if name in registry:
				prt.info(f'A {msg_name} has already been registered in {self}, now overwriting')
			else:
				prt.debug(f'Registering {msg_name} in {self}')
		
		registry.new(name, **info)
	
	def register_script(self, name, fn, description=None, use_config=False):
		'''
		Function to register a script

		:param name: name of script
		:param fn: script function (usually a callable that expects the config object)
		:param use_config: :code:`True` if the config should be passed as only arg when calling the script function, otherise it will automatically pull all arguments in the script function signature
		:param description: a short description of what the script does
		:return:
		'''
		self.register_artifact('script', name, info=dict(fn=fn, description=description,
		                                                 use_config=use_config, project=self))
	
	def register_component(self, name, fn, description=None):
		'''
		`fn` takes a single input - a Config object
		The config object is guaranteed to have at least one entry with key "_type" and the value is the same as
		the registered name of the component.

		:param name: str (should be unique)
		:param fn: callable accepting one arg (a Config object) (these should usually be classes)
		:param description: description of what this component is about
		'''
		self.register_artifact('component', name, info=dict(fn=fn, description=description,
		                                                    project=self))
	
	def register_modifier(self, name, fn, description=None, expects_config=False):
		'''
		`fn` takes as input a component and a Config object.

		:param name: str (should be unique)
		:param fn: callable accepting one arg (the "create_fn" of a registered component) (these should usually be classes)
		:param description: description of what this modifier is about
		'''
		self.register_artifact('modifier', name, info=dict(fn=fn, description=description,
		                                                   expects_config=expects_config, project=self))
	
	def register_config(self, name, path):
		'''
		Register a file as a named config

		:param name: str (should be unique)
		:param path: full path to the config
		'''
		self.register_artifact('config', name, info=dict(path=path, project=self))
	
	def register_config_dir(self, path, recursive=False, prefix=None, joiner='/'):
		'''
		Registers all yaml files found in the given directory (possibly recursively)

		When recusively checking all directories inside, the internal folder hierarchy is preserved
		in the name of the config registered, so for example if the given ``path`` points to a
		directory that contains a directory ``a`` and two files ``f1.yaml`` and ``f2.yaml``:

		Contents of ``path`` and corresponding registered names:

			- ``f1.yaml`` => ``f1``
			- ``f2.yaml`` => ``f2``
			- ``a/f3.yaml`` => ``a/f3``
			- ``a/b/f4.yaml`` => ``a/b/f3``

		If a ``prefix`` is provided, it is appended to the beginning of the registered names

		:param path: path to root directory to search through
		:param recursive: search recursively through sub-directories for more config yaml files
		:param prefix: prefix for names of configs found herein
		:param joiner: string to merge directories when recursively searching (default ``/``)
		:return: None
		'''
		assert os.path.isdir(path), f'invalid dir: {path}'
		for fname in os.listdir(path):
			parts = fname.split('.')
			candidate = os.path.join(path, fname)
			if os.path.isfile(candidate) and len(parts) > 1 and parts[-1] in {'yml', 'yaml'}:
				name = parts[0]
				if prefix is not None:
					name = joiner.join([prefix, name])
				self.register_config(name, os.path.join(path, fname))
			elif recursive and os.path.isdir(candidate):
				newprefix = fname if prefix is None else joiner.join([prefix, fname])
				self.register_config_dir(candidate, recursive=recursive, prefix=newprefix, joiner=joiner)
	
	# endregion
	
	# region Artifacts
	
	def has_artifact(self, atype, name):
		'''Check if this workspace has an artifact of type `atype` registered with name `name`'''
		registry = self._registries.get(atype, None)
		if registry is None:
			raise UnknownArtifactError(atype)
		return name in registry
	
	def find_artifact(self, atype, name):
		'''
		Find a registered artifact from the type and name
		
		:param atype: component, modifier, script, or config
		:param name: registered name
		:return: artifact entry, throws UnknownArtifactError if `atype` is not recognized,
		and the corresponding artifact missing error if the name cannot be found
		'''
		registry = self._registries.get(atype, None)
		if registry is None:
			raise UnknownArtifactError(atype)
		
		artifact = registry.get(name, None)
		if artifact is None:
			assert atype in artifact_errors, f'Missing error for {atype}'
			raise artifact_errors[atype](name)
		
		return artifact

	def view_artifacts(self, atype):
		'''Return a shallow copy of the full registry of type `atype`'''

		registry = self._registries.get(atype, None)
		if registry is None:
			raise UnknownArtifactError(atype)

		return registry.copy()

	def has_script(self, name):
		return self.has_artifact('script', name)
	def find_script(self, name):
		return self.find_artifact('script', name)
	def view_scripts(self):
		return self.view_artifacts('script')
	
	def has_component(self, name):
		return self.has_artifact('component', name)
	def find_component(self, name):
		return self.find_artifact('component', name)
	def view_components(self):
		return self.view_artifacts('component')
	
	def has_modifier(self, name):
		return self.has_artifact('modifier', name)
	def find_modifier(self, name):
		return self.find_artifact('modifier', name)
	def view_modifiers(self):
		return self.view_artifacts('modifier')
	
	def has_config(self, name):
		if self.has_artifact('config', name):
			return True
		path = Path(name)
		return path.suffix in {'.yaml', '.yml'} and path.is_file()
	def find_config(self, name):
		try:
			entry = self.find_artifact('config', name)
		except MissingConfigError:
			path = Path(name)
			if path.suffix in {'.yaml', '.yml'} and path.is_file():
				return Config_Registry.default(str(path))
				return str(path)
			
			if not path.is_dir():
				root = os.environ.get('FOUNDATION_SAVE_DIR', None)
				if root is not None:
					npath = root / path
					if npath.is_dir():
						path = npath
			
			if path.is_dir():
				cpath = path / 'config.yaml'
				if cpath.is_file():
					return Config_Registry.default(str(cpath))
					return str(cpath)
				cpath = path / 'config.yml'
				if cpath.is_file():
					return Config_Registry.default(str(cpath))
					return str(cpath)
			
			raise
		return entry
	def view_configs(self):
		return self.view_artifacts('config')
	
	# endregion
	
	# region Create/Run
	
	def run(self, script_name=None, config=None, **meta_args):
		'''
		This actually runs the script given the ``config`` object.

		Before starting the script, all meta rules are executed in order of priority (low to high)
		as they may change the config or script behavior, then the run mode is created, which is
		then called to execute the script specified in the config object (or manually overridden
		using ``script_name``)

		:param script_name: registered script name to run (overrides what is specified in ``config``)
		:param config: config object (usually created with :func:`get_config()` (see :ref:`config:Config System`)
		:param meta_args: Any additional meta arguments to include before running
		:return: script output
		'''
		if config is None:
			config = self.create_config()
		else:
			config.set_project(self)
		
		if script_name is not None:
			config.push('_meta.script_name', script_name, overwrite=True, silent=True)
		for k, v in meta_args.items():
			config.push(('_meta', k), v, overwrite=True, silent=True)
		# config._meta.update(meta_args)
		
		for rule in self.meta_rules_fns():
			config = rule(config.sub('_meta'), config)
		
		config.push('_meta._type', 'run_mode/default', overwrite=False, silent=True)
		silent = config.pull('_meta._quiet_run_mode', True, silent=True)
		mode = config.pull('_meta', silent=silent)
		# config = mode.process(config)
		
		return mode.run(config.sub('_meta'), config)
	
	def create_component(self, info):
		'''
		Creates the component specified in info (checks component registry using info.pull('_type'),
		and modifier registry for info.pull('_mod'))

		_mod can be a list, in which case they will be applied in the given order, eg:

		let mods = [A, B, C]

		component <- C(B(A(component)))

		_mod can also be a dict, in which case the keys should be the mod names and the values the order (low to high).
		So for the same behavior as above, a _mod could also be {A:0, B:1, C:2}

		NOTE: generally, modifiers should be ordered from more specific to more general

		:param info: should be a config object with attribute "_type" (and optionally "_mod")
		:return: component created using the provided config (``info``)
		'''
		
		name = info.pull('_type', silent=True)
		
		component = self.find_component(name).fn
		
		allow_missing_mods = info.pull('allow-missing-mods', False, silent=True)
		
		mod_names = info.pull('_mod', None, silent=True)
		if mod_names is not None and len(mod_names):
			
			if isinstance(mod_names, dict):
				mod_names = sorted(mod_names.keys(), key=lambda k: mod_names[k])
			
			if not isinstance(mod_names, (list, tuple)):
				mod_names = mod_names,
			
			for mod_name in mod_names:  # WARNING: apply modifications in reverse order
				
				try:
					mod = self.find_modifier(mod_name)
				except MissingModifierError:
					if allow_missing_mods:
						prt.error(f'Could not find mod "{mod_name}" for component "{name}"')
						continue
					else:
						raise
				component = mod.fn(component, info) if mod.expects_config else mod.fn(component)

		args = info.pull('__args', (), silent=True, raw=True)
		kwargs = info.pull('__kwargs', {}, silent=True, raw=True)
		
		obj = component(info, *args, **kwargs)
		if isinstance(obj, Cerifiable):
			obj.__certify__(info)
		return obj
	
	# endregion
	
	# region Config
	
	def process_argv(self, argv=(), script_name=None):
		'''
		Parses the command line arguments to identify the meta arguments, script name
		(optionally overridden using ``script_name``), and config args

		From that, this builds the config and meta config object.

		:param argv: list of all command line arguments to parse in order
		:param script_name: optional script name to override any script specified in ``argv``
		:return: config object (containing meta config under ``_meta``)
		'''
		
		# check for meta args and script name
		
		meta = {}
		
		waiting_key = None
		waiting_meta = 0
		
		remaining = []
		for i, arg in enumerate(argv):
			
			if waiting_meta > 0:
				if waiting_key in meta and isinstance(meta[waiting_key], list):
					meta[waiting_key].append(parse_arg(arg))
				else:
					meta[waiting_key] = parse_arg(arg)
				waiting_meta -= 1
				if waiting_meta == 0:
					waiting_key = None
			
			elif arg.startswith('-') and not arg.startswith('--'):
				text = arg[1:]
				for rule in self.meta_rules():
					name = rule.name
					code = rule.code
					if code is not None and text.startswith(code):
						text = text[len(code):]
						num = rule.num_args
						if num:
							if len(text):
								raise AmbiguousRuleError(code, text)
							waiting_key = name
							waiting_meta = num
							if num > 1:
								meta[waiting_key] = []
						else:
							meta[name] = True
					if not len(text):
						break
			
			elif arg == '_' or script_name is not None:
				remaining = argv[i + int(script_name is None):]
				break
			
			else:
				script_name = arg
		
		if script_name is not None:
			meta['script_name'] = script_name
		
		# create config with remaining argv
		config = self.create_config(*remaining)
		config.sub('_meta').update(meta)
		
		return config
	
	def _load_config_from_path(self, path, process=True):
		'''
		Load the yaml file and transform data to a config object

		Generally, ``get_config`` should be used instead of this method

		:param path: must be the full path to a yaml file
		:param process: if False, the loaded yaml data is passed without converting to a config object
		:return: loaded data from path (usually as a config object)
		'''
		# path = find_config_path(path)
		data = load_yaml(path)  # TODO setup global enable other file types
		
		if data is None:
			data = {}
		
		if process:
			return configurize(data)
		return data
	
	def _merge_configs(self, configs, parent_defaults=True):
		'''
		configs should be ordered from oldest to newest (ie. parents first, children last)

		This is an internal method used by ``get_config()`` and should generally not be called manually.
		'''
		
		if not len(configs):
			return self.create_config()
		
		child = configs.pop()
		merged = self._merge_configs(configs, parent_defaults=parent_defaults)
		
		# load = child.load if 'load' in child else None
		merged.update(child)
		
		return merged
	
	def _process_single_config(self, data, process=True, parents=None, tree=None, me=''):
		'''
		This loads the data (if a path or name is provided) and then checks for parents and loads those as well

		Generally, ``get_config`` should be used instead of this method

		:param data: config name or path or raw data (dict/list) or config object
		:param process: configurize loaded data
		:param parents: if None, no parents are loaded, otherwise it must be a dict where the keys are the absolute paths to the config (yaml) file and values are the loaded data
		:return: loaded data (as a config object or raw)
		'''
		
		if isinstance(data, str):
			data = self.find_config(data)
		if isinstance(data, Config_Registry.entry_cls):
			me = data
			data = self._load_config_from_path(data.path, process=process)
		
		if parents is not None and 'parents' in data:
			todo = []
			if tree is not None:# and me not in edges:
				tree[me] = []
			for parent in data['parents']:  # prep new parents
				# ppath = _config_registry[parent] if parent in _config_registry else parent
				entry = self.find_config(parent)
				# ppath = entry.path
				if entry.path not in parents:
					todo.append(entry)
					parents[entry] = None
				if tree is not None:
					tree[me].append(entry)
			for entry in todo:  # load parents
				proj = self if entry.project is None else entry.project
				parents[entry] = proj._process_single_config(entry, parents=parents, tree=tree)
		elif tree is not None:
			tree[me] = []
		
		return data
	
	def create_config(self, *contents, **parameters):  # Top level function
		'''
		Top level function for users. This is the best way to load/create a config object.

		All parent config (registered names or paths) that should be loaded
		must precede any manual entries, and will be loaded in reverse order (like python class inheritance).

		If the key ``_history_key`` is specified and not :code:`None`, a flattened list of all parents of
		this config is pushed to the given key.

		:param contents: registered configs or paths or manual entries (like in terminal)
		:param parameters: specify parameters manually as key value pairs
		:return: config object
		'''
		root = configurize({})
		if len(contents) + len(parameters) == 0:
			root.set_project(self)
			return root
		
		reg = []
		terms = {**parameters}
		allow_reg = True
		waiting_key = None
		
		for term in contents:
			
			if term.startswith('--'):
				allow_reg = False
				if waiting_key is not None:
					terms[waiting_key] = True
				waiting_key = term[2:]
			
			elif waiting_key is not None:
				terms[waiting_key] = parse_arg(term)
				waiting_key = None
			
			elif allow_reg:
				reg.append(term)
			
			else:
				raise Exception(f'Parsing error: {term} in {contents}')
		
		if waiting_key is not None:
			terms[waiting_key] = True
		
		root.update(configurize(terms))
		
		if len(reg) == 0:
			root.set_project(self)
		
			return root
		
		root['parents'] = configurize(data=reg + (list(root['parents']) if 'parents' in root else []))
		
		parents = {}
		tree = {}
		
		root_id = ''
		root = self._process_single_config(root, parents=parents, tree=tree, me=root_id)
		
		pnames = []
		if len(parents):  # topo sort parents
			
			order = linearize(tree, heads=[root_id], order=True)[root_id]
			
			pnames = [(node.name if node.project is None or not isinstance(node.project, Project)
			           else f'{node.project}:{node.name}') for node in order[1:]]
			order = [root] + [parents[p] for p in order[1:]]
			
			# for analysis, record the history of all loaded parents
			order = list(reversed(order))
		
		else:
			order = [root]
		
		root = self._merge_configs(order, )
		root.set_project(self)
		
		include_history = root.pull('_ancestry_key', 'ancestors', silent=True)
		if include_history is not None:
			root.push(include_history, pnames, silent=True)
		
		root.push('parents', '_x_', silent=True)
		
		return root
	
	# endregion
	
	pass


class Project(Workspace):
	'''
	Projects are used to group code into packets with specific config files that should be
	loaded all together. A project must contain a yaml file named ``.fig.yml`` in the
	root directory of the project (aka the "project directory"), and all paths in
	that yaml file should be relative to the project directory.

	Generally there are two kinds of projects: "packages" and "loose" projects.
	Package projects are meant to be installed and used as a library, while loose
	projects may just be a series of python files.

	This class may also be subclassed to change the behavior of projects (such as changing the loading),
	in fact, any subclasses of this class can automatically be registered when providing a name for the
	project type in the class definition.
	'''
	
	# required_attrs = ['name', 'author', 'info_path']
	# recommended_attrs = ['url', 'version', 'license', 'description']
	
	def __init_subclass__(cls, name=None):
		'''Subclasses can automatically be registered if a ``ptype`` for the registry is provided'''
		cls.ptype = name
		register_project_type(name, cls)
	
	def __init__(self, profile=None, **kwargs):
		super().__init__(**kwargs)
		
		self._profile = profile
	
	# region Getters
	
	def get_profile(self):
		return self._profile
	
	def get_path(self):
		'''Gets the path to the project directory'''
		return self.root
	
	def get_related(self):
		'''Returns a list of project names of all projects that should be loaded prior to this one'''
		return self.related
	
	# endregion
	
	@staticmethod
	def check_project_type(raw):
		'''
		Based on raw project info, check if this project expects a certain project type, if so
		optionally provide a path to the source file of the project type.

		:param raw: raw project info
		:return: None or tuple with project type identifier and path to source file (or None)
		'''
		ptype = raw.get('project_type', None)
		if ptype is not None:
			return ptype, raw.get('ptype_src_file', None)
	
	def _process(self, raw):
		'''
		Given the raw info loaded from a yaml file, this function checks integrates the information into
		the project object (self).

		:param raw: dictionary of info (usually loaded from a yaml)
		'''
		
		# region path
		
		self.info_path = raw.get('info_path', None)
		self.root = Path(self.info_path).parents[0]
		
		# endregion
		
		if 'py_info' in raw:
			info = {'__file__': str(self.root / raw['py_info'])}
			with open(info['__file__'], 'r') as f:
				exec(f.read(), info)
			del info['__file__']
			raw.update(info)
			info.update(raw)
			raw = info
		
		# region info
		
		self.name = raw.get('name', None)
		
		author = raw.get('author', None)
		self.primary_author = raw.get('primary_author', author)
		self.authors = raw.get('authors', None)
		if self.authors is None and self.primary_author is not None:
			self.authors = [self.primary_author]
		if self.primary_author is None and self.authors is not None and len(self.authors):
			self.primary_author = self.authors[0]
		
		self.github = raw.get('github', None)
		self.url = raw.get('url', None)
		if self.github is not None and self.url is None:
			self.url = f'https://github.com/{self.github}'
		
		self.version = raw.get('version', None)
		self.license = raw.get('license', None)
		# self.use = raw.get('use', 'leaf')  # {leaf, package}
		
		self.add_to_path = raw.get('add_to_path', True)
		
		self.description = raw.get('description', None)
		if self.description is not None:
			self.__doc__ = self.description
		
		for key in raw:
			if key not in self.__dict__:
				prt.info(f'Found optional project info: {key}')
				setattr(self, key, raw[key])
		
		
		# endregion
		
		super()._process(raw)
		
		# region related
		
		self.related = raw.get('related', [])  # should be a list of idents (paths or names in profile)
		# self.dependenies = raw.get('dependency', []) # names of projects that must be loaded before this one
		
		# endregion
		# region components
		
		# self.last_update = raw.get('last_update', None)
		self.conda_env = raw.get('conda', None)  # TODO
		
		self.package = raw.get('package', None)
		
		self.no_auto_config = raw.get('no_auto_config', None)
		if not self.no_auto_config:
			for aname in ['config', 'configs']:
				path = self.root / aname
				if path.is_dir() and str(path) not in self.config_paths:
					self.config_paths.append(str(path))
		
		# endregion
		
		prt.debug(f'Finished loading project info for {self.name} ({self.info_path})')
	
	def initialize(self):
		root = self.get_path()
		
		if self.add_to_path:
			sys.path.append(str(root))
		
		origin = os.getcwd()
		os.chdir(str(root))
		
		super().initialize()
		
		name = self.get_name()
		if self.package is not None and name not in sys.modules:
			path = root / self.package / '__init__.py'
			if path.exists():
				spec = importlib.util.spec_from_file_location(name, str(path))
				mod = importlib.util.module_from_spec(spec)
				sys.modules[mod.__name__] = mod
				projects_module.__dict__[mod.__name__] = mod
				spec.loader.exec_module(mod)
				# mod.MyClass()
			
		os.chdir(origin)
		
		prt.info(f'Project "{name}" initialized')
	
	# region Registration
	
	def register_artifact(self, atype, name, info, include_global=True):
		info['project'] = self
		super().register_artifact(atype, name=name, info=info)
		if include_global:
			self.get_profile().register_artifact(atype, name=name, info=info)
	
	# endregion
	
	# region Artifacts
	
	def has_artifact(self, atype, name, check_global=True):
		
		if ':' in name:
			return self.get_profile().has_artifact(atype, name)
		
		if super().has_artifact(atype, name):
			return True
		return not check_global or self.get_profile().has_artifact(atype, name)
	
	def find_artifact(self, atype, name, check_global=True):
		
		if ':' in name:
			return self.get_profile().find_artifact(atype, name)
		
		try:
			artifact = super().find_artifact(atype, name)
		except MissingArtifactError:
			if not check_global:
				raise
			artifact = self.get_profile().find_artifact(atype, name)
		
		return artifact
	
	def has_config(self, name):
		if super().has_config(name):
			return True
		path = self.get_path() / Path(name)
		return path.suffix in {'.yaml', '.yml'} and path.is_file()
	
	def find_config(self, name):
		try:
			path = super().find_config(name)
		except MissingConfigError:
			raise
			path = self.get_path() / Path(name)
			if path.suffix in {'.yaml', '.yml'} and path.is_file():
				return str(path)
			raise
		else:
			return path
	
	def view_artifacts(self, atype):
		glob = self.get_profile().view_artifacts(atype)
		glob.update(super().view_artifacts(atype))
		return glob


# endregion


register_project_type('default', Project)







import sys, os
import yaml

from omnibelt import get_printer, get_now, save_yaml, load_yaml

from .errors import WrongInfoContainerType

prt = get_printer(__name__)



class Container:
	'''
	Keeps track of some data extracted from a yaml.
	Consequently a manager is entirely defined by a path to the yaml file
	which holds the source data, however managers can be subclassed to define
	custom behavior in how that data is used.
	
	Main uses include projects and profiles.
	
	Generally any attribute holds data that can be exported,
	however attributes that start with "_" are skipped
	
	'''
	required_attrs = []
	recommended_attrs = []
	
	def __init__(self, name=None, path=None, raw=None):
		self._updated = False
		
		self.name = name
		
		if path is not None:
			raw = self.load_raw_info(path)
			
		self.process(raw)
		
	
	@staticmethod
	def find_container_type(raw):
		'''
		Based on the raw info, confirm that this is the right type to use as the manager.
		This enables using custom subclasses of Manager
		
		:param raw: src info (usually from some yaml file)
		:return: must throw a WrongManager exception if a different one should be used
		'''
		
		ctype = raw.get('type', None)
		
		if ctype is not None:
			ctype_src = raw.get('type_src', None)
			raise WrongInfoContainerType(ctype, ctype_src)
		
	@staticmethod
	def load_raw_info(path):
		'''Loads the info yaml file'''
		raw = load_yaml(path) if os.path.isfile(path) else None
		if raw is None:
			raw = {}
		raw['info_path'] = path # automatically set info_path to the path
		raw['info_dir'] = os.path.dirname(path)
		return raw
	
	# region Getters
	
	def get_name(self):
		'''Gets the container name'''
		return self.name
	def get_info_path(self):
		'''Gets the path to the info file'''
		return getattr(self, 'info_path', None)
	
	# endregion
	
	def __str__(self):
		return f'{self.get_name()}'
	
	def __repr__(self):
		return f'{self.__class__.__name__}({self.get_name()})'
	
	
	def _process(self, raw):
		'''
		Extracts/formats information from the loaded info file.
		
		:param raw: loaded yaml file with contents
		:return:
		'''
		pass
	
	def process(self, raw=None):
		'''
		Extracts/formats information from the loaded info file.
		Calls `_process()` which should be overridden by subclasses.
		
		:param raw: loaded yaml file with contents
		:return:
		'''
		if raw is None:
			raw = {}
		
		self.find_container_type(raw)
		
		self._process(raw)
		
		for key in raw:
			if key not in self.__dict__:
				setattr(self, key, raw[key])
		
		for key in self.required_attrs:
			if getattr(self, key) is None:
				prt.error(f'{key} not found in {self.__class__.__name__} {self}')

		for key in self.recommended_attrs:
			if getattr(self, key) is None:
				prt.warning(f'{key} not found in {self.__class__.__name__} {self}')
	
	
	
	def export(self, path=None):
		'''
		Saves info to yaml (by default where it was loaded from)

		Filters out any entries with keys that start with '_' or have no value (None)
		
		:return: path to export if one is provided, otherwise all the data that would have been exported as a dict
		'''
		
		if path is None:
			path = self.get_info_path()
		
		data = {}
		for k, v in self.__dict__.items():
			if v is not None and k[0] != '_':
				data[k] = v
		
		if self._updated: # something was changed
			data['last_update'] = get_now()
		
		if path is None:
			prt.warning('No export path found, so there is nowhere to save the info')
			return
		
		save_yaml(data, path)
		
		prt.debug(f'Exported {repr(self)} to {path}')
		return path
	
	def cleanup(self):
		if self._updated:
			self.export()
	




from .loading import get_profile

from .decorators import AutoScript, Script, Component, AutoComponent, Modifier, AutoModifier, Modification

from .config import ConfigIter, EmptyElement, ConfigType
Component('iter')(ConfigIter)
del ConfigIter

from .top import get_project, get_current_project, \
	entry, main, run, quick_run, initialize, cleanup, \
	get_config, create_component, quick_create, \
	register_script, register_config, register_component, \
	register_config_dir, register_modifier, resolve_order, \
	find_script, find_component, find_modifier, find_config, \
	has_script, has_component, has_config, has_modifier, \
	view_configs, view_components, view_scripts, view_modifiers

from .rules import Meta_Rule
from .modes import Run_Mode
from . import common
from .common import Configurable
from.organization import Cerifiable
# from .farming import Farmer, Worker

from omnibelt import AbortTransaction

import os
__info__ = {'__file__':os.path.join(os.path.abspath(os.path.dirname(__file__)), '_info.py')}
with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), '_info.py'), 'r') as f:
	exec(f.read(), __info__)
del os
del __info__['__file__']
__author__ = __info__['author']
__version__ = __info__['version']




import sys, traceback

from omnibelt import unspecified_argument, InitWall

import multiprocessing

from .decorators import Component
from .config import configurize, pythonize, ConfigType
from .top import get_config, run
from .common import Configurable


class ExceptionWrapper(object):
	r"""Wraps an exception plus traceback to communicate across threads"""
	def __init__(self, exc_info):
		# It is important that we don't store exc_info, see
		# NOTE [ Python Traceback Reference Cycle Problem ]
		self.exc_type = exc_info[0]
		self.exc_msg = "".join(traceback.format_exception(*exc_info))


def _fmt_worker_config(config, ID=None):
	config = configurize(config)
	config.push('worker_ID', ID, silent=True)
	config.set_process_id(ID)
	return config


def _worker_loop(ID, in_q, out_q, config):
	
	config = _fmt_worker_config(config, ID=ID)
	
	try:
		worker = config.pull_self()
	except Exception:
		print(f'** Worker {ID} failed **')
		out_q.put(ExceptionWrapper(sys.exc_info()))
		return
	
	while True:
		config = in_q.get()

		if config is None:
			break
		try:
			config = _fmt_worker_config(config, ID=ID)
			out = worker(config)
		except Exception:
			out_q.put(ExceptionWrapper(sys.exc_info()))
		else:
			out_q.put(out)


class Worker(Configurable):
	def __call__(self, config):
		raise NotImplementedError


@Component('worker')
class ScriptWorker(Worker):
	def __init__(self, A, **kwargs):
		
		init_script = A.pull('init_script', None)
		script = A.pull('script')
		
		update_config = A.pull('update_config', False)
		
		super().__init__(A, **kwargs)
		
		self.script = script
		
		if init_script is not None:
			out = run(init_script, A)
			replace_config = A.pull('replace_config', out is not None)
			if isinstance(out, ConfigType) and replace_config:
				A = out
		
		self.config = A if update_config else None
		
	def __call__(self, config):
		if self.config is not None and isinstance(config, ConfigType):
			self.config.update(config)
			config = self.config
		return run(self.script, config)



class FarmerBase(InitWall):
	def __init__(self, worker_script=None, worker_init_script=None, worker_config=None,
	             update_worker_config=None, waiting=None, auto_dispatch=True,
	             num_workers=0, timeout=20, mp_lib=None, **other):
		
		if worker_config is None:
			assert worker_script is not None, 'no worker script or config provided'
			worker_config = get_config()
		
		worker_config.push('script', worker_script, overwrite=False, silent=True)
		if worker_init_script is not None:
			worker_config.push('init_script', worker_init_script, overwrite=False, silent=True)
		if update_worker_config is not None:
			worker_config.push('update_worker_config', update_worker_config, overwrite=False, silent=True)
		
		super().__init__(**other)
		
		worker_config.push('_type', 'worker', silent=True, overwrite=False)
		
		if mp_lib is None:
			mp_lib = multiprocessing
		self.mp_lib = mp_lib
		
		self.worker_config = worker_config
		
		self.waiting = waiting
		self.auto_dispatch = auto_dispatch
		
		self.num_workers = num_workers
		self.timeout = timeout
		
		self.worker = None
		self.workers = None
		
		self.start()
	
	
	def start(self, num_workers=None):
		
		if num_workers is None:
			num_workers = self.num_workers
		
		if self.workers is not None:
			self.stop()
		
		self.in_queue = self.mp_lib.Queue()
		self.out_queue = self.mp_lib.Queue()
		
		if num_workers > 0:
			self.workers = [
				self.mp_lib.Process(target=_worker_loop, args=(i, self.in_queue, self.out_queue, self.worker_config))
				for i in range(num_workers)]
			
			for w in self.workers:
				w.daemon = True  # ensure that the worker exits on process exit
				w.start()
			
			if self.waiting is None:
				self.waiting = num_workers if self.auto_dispatch else 0
			self.send(self.waiting)
		
		else:
			self.worker = self.worker_config.pull_self()
		
		
	def stop(self):
		
		self.in_queue.close()
		self.in_queue = None
		self.out_queue.close()
		self.out_queue = None
		
		if self.worker is not None:
			self.worker = None
		else:
			for _ in self.workers:
				self.dispatch(None)
			
			for worker in self.workers:
				worker.stop()
	
	
	def dispatch(self, config=unspecified_argument):
		if config is unspecified_argument:
			config = self._next_job()
		
		if self.worker is None:
			self.in_queue.put(pythonize(config))
		else:
			self.out_queue.put(self.worker(config))
	
	
	def quick_dispatch(self, *args, **kwargs):
		return self.dispatch(get_config(*args, **kwargs) if len(args) or len(kwargs) else None)
	
	
	def send(self, num=None, block=False, timeout=unspecified_argument):
		if num is None or num == 0:
			self.dispatch()
		else:
			for _ in range(num):
				self.dispatch()
		
		if block:
			return self.receive(num=num, timeout=timeout)
	
	
	def receive(self, num=None, timeout=unspecified_argument):
		if timeout is unspecified_argument:
			timeout = self.timeout
		
		if num is None or num == 0:
			return self.out_queue.get(block=True, timeout=timeout)
		return [self.out_queue.get(block=True, timeout=timeout) for _ in range(num)]
	
	def complete(self):
		return self.send(len(self), block=True)
	
	def _next_job(self):
		'''creates config object to send to the worker, or throws StopIteration if there are no more jobs'''
		raise NotImplementedError
	
	def __len__(self):
		raise NotImplementedError
	
	def __iter__(self):
		return self
	
	
	def __next__(self):
		return self.send(block=True)



class Farmer(Configurable, FarmerBase):
	def __init__(self, A, worker_script=unspecified_argument, worker_init_script=unspecified_argument,
	             worker_config=unspecified_argument,
	             update_worker_config=None, waiting=unspecified_argument, auto_dispatch=None,
	             num_workers=None, timeout=unspecified_argument, **kwargs):
		
		if worker_script is unspecified_argument:
			worker_script = A.pull('script', None)
		
		if worker_init_script is unspecified_argument:
			worker_init_script = A.pull('init-script', None)
			
		if worker_config is unspecified_argument:
			worker_config = A.pull('worker-config', '<>worker', None, raw=True)
		
		if update_worker_config is None:
			update_worker_config = A.pull('update_worker_config', False)
			
		if waiting is unspecified_argument:
			waiting = A.pull('waiting', None)
		
		if auto_dispatch is None:
			auto_dispatch = A.pull('auto-dispatch', False)
			
		if num_workers is None:
			num_workers = A.pull('num-workers', 0)
			
		if timeout is unspecified_argument:
			timeout = A.pull('timeout', 20)
		
		super().__init__(A, worker_script=worker_script, worker_init_script=worker_init_script,
		                 worker_config=worker_config,
	             update_worker_config=update_worker_config, waiting=waiting, auto_dispatch=auto_dispatch,
	             num_workers=num_workers, timeout=timeout, **kwargs)










# from omnibelt import monkey_patch

from .top import register_script, register_component, register_modifier
from .util import autofill_args

def Script(name, description=None, use_config=True):
	'''
	Decorator to register a script

	:param name: name of script
	:param description: a short description of what the script does
	:param use_config: :code:`True` if the config should be passed as only arg when calling the script function, otherise it will automatically pull all arguments in the script function signature
	:return: decorator function expecting a callable
	'''
	def _reg_script_decorator(fn):
		nonlocal name, use_config
		register_script(name, fn, use_config=use_config, description=description)
		return fn
	# monkey_patch(_reg_script_decorator)
	
	return _reg_script_decorator

def AutoScript(name, description=None):
	'''
	Convienence decorator to register scripts that automatically extract
	relevant arguments from the config object

	:param name: name of the script
	:param description: a short description of what the script does
	:return: decorator function expecting a callable that does not expect the config as argument (otherwise use :func:`Script`)
	'''
	return Script(name, use_config=False, description=description)


def Component(name=None):
	'''
	Decorator to register a component

	NOTE: components should usually be types/classes to allow modifications

	:param name: if not provided, will use the __name__ attribute.
	:return: decorator function
	'''
	
	def _register_cmp_decorator(cmp):
		nonlocal name
		if name is None:
			name = cmp.__name__
		register_component(name, cmp)
		return cmp
	# monkey_patch(_register_cmp_decorator)
	
	return _register_cmp_decorator


def AutoComponent(name=None, aliases=None, auto_name=True):
	'''
	Instead of directly passing the config to an AutoComponent, the necessary args are auto filled and passed in.
	This means AutoComponents are somewhat limited in that they cannot modify the config object and they cannot be
	modified with AutoModifiers.

	Note: AutoComponents are usually components that are created with functions (rather than classes) since they can't
	be automodified. When registering classes as components, you should probably use `Component` instead, and pull
	from the config directly.

	:param name: name to use when registering the auto component
	:param aliases: optional aliases for arguments used when autofilling (should be a dict[name,list[aliases]])
	:return: decorator function
	'''
	
	def _auto_cmp_decorator(cmp):
		nonlocal name, aliases
		
		if type(cmp) == type:  # to allow AutoModifiers
			
			cls = type(f'Auto_{cmp.__name__}' if auto_name else cmp.__name__, (cmp,), {})
			# monkey_patch(cls)
			
			def _cmp_init_fn(self, info):
				args, kwargs = autofill_args(cmp, info, aliases=aliases, run=False)
				super(cls, self).__init__(*args, **kwargs)
			# monkey_patch(_cmp_init_fn)
			
			cls.__init__ = _cmp_init_fn
			
			_auto_create_fn = cls
		
		else:
			def _auto_create_fn(config):
				nonlocal cmp, aliases
				return autofill_args(cmp, config, aliases)
			
			# monkey_patch(_auto_create_fn)
		Component(name)(_auto_create_fn)
		
		return cmp
	
	# monkey_patch(_auto_cmp_decorator)
	
	return _auto_cmp_decorator


def Modifier(name=None, expects_config=False):
	'''
	Decorator to register a modifier

	NOTE: a :class:`Modifier` is usually not a type/class, but rather a function
	(except :class:`AutoModifiers`, see below)

	:param name: if not provided, will use the __name__ attribute.
	:param expects_config: True iff this modifier expects to be given the config as second arg
	:return: decorator function
	'''
	
	def _mod_decorator_fn(mod):
		nonlocal name, expects_config
		if name is None:
			name = mod.__name__
		register_modifier(name, mod, expects_config=expects_config)
		return mod
	# monkey_patch(_mod_decorator_fn)
	
	return _mod_decorator_fn


def AutoModifier(name=None):
	'''
	Can be used to automatically register modifiers that combine types

	To keep component creation as clean as possible, modifier types should allow arguments to their __init__
	(other than the Config object) and only call pull on arguments not provided, that way child classes of
	the modifier types can specify defaults for the modifications without calling pull() multiple times
	on the same arg.

	Note: in a way, this converts components to modifiers (but think before using). This turns the modified
	component into a child class of this modifier and its previous type.

	In short, Modifiers are used for wrapping of components, AutoModifiers are used for subclassing components

	:param name: if not provided, will use the __name__ attribute.
	:return: decorator to decorate a class
	'''
	
	def _auto_mod_decorator(mod_type):
		nonlocal name
		
		def _the_mod_creation_fn(cmpn_type):
			# awesome python feature -> dynamic type declaration!
			cls = cmpn_type if issubclass(cmpn_type, mod_type) \
				else type('{}_{}'.format(mod_type.__name__, cmpn_type.__name__), (mod_type, cmpn_type), {})
			# monkey_patch(cls)
			return cls
		# monkey_patch(_the_mod_creation_fn)
		
		Modifier(name=mod_type.__name__ if name is None else name)(_the_mod_creation_fn)
		return mod_type
	# monkey_patch(_auto_mod_decorator)
	
	return _auto_mod_decorator


def _make_post_mod(mod):
	def _make_cmpn_decorator(cmpn):
		def _modification_fn(info):
			return mod(cmpn(info), info)
		# monkey_patch(_modification_fn)
		return _modification_fn
	# monkey_patch(_make_cmpn_decorator)
	return _make_cmpn_decorator


def Modification(name=None):
	'''
	A kind of Modifier that modifies the component after it is created,
	and then returns the modified component

	expects a callable with input (component, config)

	Modifications should almost always be applied after all other modifiers,
	so they should appear at the end of _mod list

	:param name: name to register
	:return: a decorator expecting the modification function
	'''
	
	def _reg_modification(mod):
		nonlocal name
		Modifier(name)(_make_post_mod(mod))
		return mod
	# monkey_patch(_reg_modification)
	
	return _reg_modification






import sys, os, io
import inspect
import yaml
import time

from omnibelt import primitives, get_printer, \
	get_global_settings, set_global_setting, get_global_setting, get_now, resolve_order, \
	spawn_path_options
from omnibelt.logging import global_settings as belt_global_settings

from .errors import PythonizeError, ConfigurizeFailed

LIB_PATH = os.path.dirname(__file__)

global_settings = belt_global_settings.copy()
global_settings.update({
	
	# naming settings
	'default_entry_name': 'default',
	'profile_env_var': 'FIG_PROFILE',
	'profile_src_env_var': 'FIG_PROFILE_SRC',
	'init_env_var': 'FIG_INIT',
	
	'disable_princeps': False,
	'princeps_path': 'FIG_PRINCEPS',
	
	# default infos
	'default_profile': None,
	
	'allow_profile_sources': True,
	'use_last_entry_as_default': True,
	'default_ptype': 'default',
	
	# 'resolution_order': [ # for scripts, configs, components, and modifiers
	# 	 'local', # check the current working dir if its in a project
	#      'active-project', # currently active project
	#      'related-projects', # any projects related to the active project
	#      'init' # scripts loaded through init (no project)
	#      ],
	
	'arg_parse_language': 'yaml',
	
})

prt = get_printer(__name__)



def configurize(data):
	'''
	Transform data container to use config objects (ConfigDict/ConfigList)

	:param data: dict/list data
	:return: deep copy of data using ConfigDict/ConfigList
	'''
	if isinstance(data, global_settings['config_type']):
		return data
	for typ, convert in global_settings.get('config_converters', {}).items():
		allow_subtypes = True
		if isinstance(convert, tuple):
			allow_subtypes, convert = convert
		try:
			if (allow_subtypes and isinstance(data, typ)) or type(data) == typ:
				return convert(data, configurize)
		except ConfigurizeFailed:
			pass
	return data


def pythonize(data):  # TODO: allow adding yamlify rules for custom objects
	'''
	Transform data container into regular dicts/lists to export to yaml file

	:param data: Config object
	:return: deep copy of data using dict/list
	'''
	# if data is None:
	# 	return '_None'
	if data is None or isinstance(data, primitives):
		return data
	if isinstance(data, dict):
		return {k: pythonize(v) for k, v in data.items() if not str(k).startswith('__')}
	if isinstance(data, (list, tuple, set)):
		return [pythonize(x) for x in data]
	
	raise PythonizeError(data)

# parse args

def parse_arg(arg, mode=None):
	
	if mode is None:
		mode = global_settings['arg_parse_language']
	
	if mode == 'python':
		raise NotImplementedError
	
	try:
		if isinstance(mode, str):
			if mode == 'yaml':
				return yaml.safe_load(io.StringIO(arg))
			elif mode == 'python':
				return eval(arg, {}, {})
			else:
				pass
		else:
			return mode(arg)
	except:
		pass
	return arg


# autofill

def autofill_args(fn, config, aliases=None, run=True): # TODO: move to omnibelt (?)

	params = inspect.signature(fn).parameters

	args = []
	kwargs = {}

	for n, p in params.items():

		order = [n]
		if aliases is not None and n in aliases: # include aliases
			order.extend('<>{}'.format(a) for a in aliases[n])
		if p.default != inspect._empty:
			order.append(p.default)
		elif p.kind == p.VAR_POSITIONAL:
			order.append(())
		elif p.kind == p.VAR_KEYWORD:
			order.append({})

		arg = config.pull(*order)

		if p.kind == p.POSITIONAL_ONLY:
			args.append(arg)
		elif p.kind == p.VAR_POSITIONAL:
			args.extend(arg)
		elif p.kind == p.VAR_KEYWORD:
			kwargs.update(arg)
		else:
			kwargs[n] = arg
	if run:
		return fn(*args, **kwargs)
	return args, kwargs









# region Containers

class NoValidProjectError(Exception):
	'''Raised when no project is found for the given identifier (which should be the name or path to the project)'''
	def __init__(self, ident):
		super().__init__(f'Not a valid project: {ident}')

class AmbiguousRuleError(Exception):
	def __init__(self, code, text):
		super().__init__(f'Can\'t combine multiple meta-rules if they require params: {code} in {text}')

# endregion

# region Artifacts

class UnknownArtifactError(Exception):
	pass

class MissingArtifactError(Exception):
	def __init__(self, atype, name):
		super().__init__(name)
		self.atype = atype
		self.name = name

class MissingComponentError(MissingArtifactError):
	def __init__(self, name):
		super().__init__('component', name)

class MissingModifierError(MissingArtifactError):
	def __init__(self, name):
		super().__init__('modifier', name)

class MissingConfigError(MissingArtifactError):
	def __init__(self, name):
		super().__init__('config', name)

class MissingScriptError(MissingArtifactError):
	def __init__(self, name):
		super().__init__('script', name)

artifact_errors = {'script': MissingScriptError, 'component': MissingComponentError,
                   'modifier': MissingModifierError, 'config': MissingConfigError}

# endregion

# region Config

class ConfigNotFoundError(Exception):
	'''Raised when a config parameter is not found and no viable defaults are provided'''
	
	def __init__(self, ident):
		super().__init__(f'Unknown config: {ident}')

class MissingParameterError(Exception):
	'''Raised when a config parameter was not found, and no viable defaults were provided'''
	def __init__(self, key):
		super().__init__(key)

class InvalidKeyError(Exception):
	'''Only raised when a key cannot be converted to an index for a :class:`ConfigList`'''
	pass

class UnknownActionError(Exception):
	'''Raised when trying to record an unrecognized action with the config object'''
	pass

# endregion

# region Misc


class PythonizeError(Exception):
	'''Raised when an object is unable to be turned into a yaml object (primitives, dicts, lists)'''
	def __init__(self, obj):
		super().__init__('Unable to yamlify: {} (type={})'.format(obj, type(obj)))
		self.obj = obj


class WrongInfoContainerType(Exception):
	'''Raised when trying to load a container, but the container expects a different type (ie. subclass).'''
	def __init__(self, ctype, ctype_src=None):
		super().__init__(f'Reload container using custom subclass: {ctype}')
		self.ctype = ctype
		self.ctype_src = ctype_src
	
	def get_ctype(self):
		return self.ctype
	
	def get_mtype_src(self):
		return self.ctype_src

class ConfigurizeFailed(Exception):
	'''Raised when trying to configurize an object by type, but it ends up not working'''
	pass

# endregion




import sys, os
import importlib
import importlib.util

from omnibelt import get_printer, Registry, Entry_Registry

prt = get_printer(__name__)

# region Source files

_loaded_files = {}
def include_files(*paths):
	'''
	Executes all provided paths to python files that have not already been run.
	
	:param paths: paths to python files to be executed (if not already)
	:return: None
	'''
	global _load_counter
	
	for path in paths:
		if os.path.isfile(path):
			apath = os.path.abspath(path)
			if apath not in _loaded_files:
				sys.path.insert(1, os.path.dirname(apath))
				old = os.getcwd()
				os.chdir(os.path.dirname(apath))
				prt.debug(f'Loading {apath}')
				code_block = compile(open(apath).read(), apath, 'exec')
				globs = {'__file__':apath}
				exec(code_block, globs)
				_loaded_files[apath] = globs
				
				del sys.path[1]
				os.chdir(old)

def include_package(*packages):
	'''
	Imports packages based on their names
	
	:param packages: list of package names to be imported
	:return: None
	'''
	for pkg in packages:
		importlib.import_module(pkg)

# endregion


# region project types

_ptype_registry = Registry()


def register_project_type(name, cls):
	'''
	Project types allow users to customize the behavior of project objects

	:param name: identifier of this project type
	:param cls: project type class
	:return: None
	'''

	_ptype_registry.new(name, cls)


def get_project_type(name):
	'''Gets the project type associated with that name, otherwise returns None'''
	if name not in _ptype_registry:
		prt.error(f'Project type: {name} not found')
	return _ptype_registry.get(name, None)


def view_project_types():
	'''Returns a copy of the full project type registry'''
	return _ptype_registry.copy()

# endregion


import sys
import traceback

from omnibelt import get_printer

from ..rules import Meta_Rule

CODE = 'q'
NAME = 'quiet'

prt = get_printer(__name__)


@Meta_Rule(NAME, priority=10, code=NAME, description='Set config to silent')
def quiet_rule(meta, config):
	'''
	When activated, this rule sets the config object to silent mode, which means pulls/pushes are not printed to stdout

	:param meta: meta config object
	:param config: full config object
	:return: possibly silenced config object
	'''
	
	quiet = meta.pull(NAME, False, silent=True)
	if quiet:
		config.set_silent()
	return config





from omnibelt import Entry_Registry, get_printer

prt = get_printer(__name__)

class Artifact_Registry(Entry_Registry):
	def export(self):
		return list(self.keys())



class Script_Registry(Artifact_Registry, components=['fn', 'use_config', 'description', 'project']):
	pass

class Component_Registry(Artifact_Registry, components=['fn', 'description', 'project']):
	pass

class Modifier_Registry(Artifact_Registry, components=['fn', 'description', 'expects_config', 'project']):
	pass

class Config_Registry(Artifact_Registry, components=['path', 'project']):
	@classmethod
	def default(cls, path):
		return cls.entry_cls(path, path, None)
# DefaultConfigEntry = Config_Registry.entry_cls

class Rules_Registry(Entry_Registry, components=['fn', 'priority', 'code', 'num_args', 'description']):
	pass





import sys, os

from omnibelt import get_printer, resolve_order, monkey_patch

from .external import include_files

from .util import global_settings


prt = get_printer(__name__)

# region Projects

from .loading import get_profile

def get_current_project():
	'''Get the current project, assuming a profile is loaded, otherwise returns None'''
	return get_profile().get_current_project()

def get_project(ident=None):
	'''Checks the profile to return (and possibly load) a project given the name or path ``ident``'''
	return get_profile().get_project(ident)

# endregion

# region Running


def entry(script_name=None):
	'''
	Recommended entry point when running a script from the terminal.
	This is also the entry point for the ``fig`` command.

	This collects the command line arguments in ``sys.argv`` and overrides the
	given script with ``script_name`` if it is provided

	:param script_name: script to be run (may be set with arguments) (overrides other arguments if provided)
	:return: None
	'''
	argv = sys.argv[1:]
	main(*argv, script_name=script_name)


def main(*argv, script_name=None):
	'''
	Runs the desired script using the provided ``argv`` which are treated as command line arguments

	Before running the script, this function initializes ``omni-fig`` using :func:`initialize()`,
	and then cleans up after running using :func:`cleanup()`.

	:param argv: raw arguments as if passed in through the terminal
	:param script_name: name of registered script to be run (may be set with arguments) (overrides other arguments if provided)
	:return: output of script that is run
	'''
	
	initialize()
	
	project = get_current_project()
	
	config = project.process_argv(argv, script_name=script_name)
	out = project.run(config=config)
	
	cleanup()
	
	return out


def run(script_name, config, **meta):
	'''
	Runs the specified script registered with ``script_name`` using the current project.
	
	:param script_name: must be registered in the current project or defaults to the profile
	:param config: config object passed to the script
	:param meta: any meta rules that modify the way the script is run
	:return: output of the script, raises MissingScriptError if the script is not found
	'''
	return get_current_project().run(script_name=script_name, config=config, **meta)


def quick_run(script_name, *parents, **args):
	'''
	Convenience function to run a simple script without a given config object,
	instead the config is entirely created using the provided ``parents`` and ``args``.

	:param script_name: name of registered script that is to be run
	:param parents: any names of registered configs to load
	:param args: any additional arguments to be provided manually
	:return: script output
	'''
	config = get_config(*parents, **args)
	return run(script_name, config)


def initialize(*projects, **overrides):
	'''
	Initializes omni-fig by running the "princeps" file (if one exists),
	loading the profile, and any active projects. Additionally loads the
	project in the current working directory (by default).

	Generally, this function should be run before running any scripts, as it should register all
	necessary scripts, components, and configs when loading a project. It is automatically called
	when running the :func:`main()` function (ie. running through the terminal). However, when
	starting scripts from other environments (such as in a jupyter notebook), this should be called
	manually after importing ``omnifig``.

	:param projects: additional projects that should be initialized
	:param overrides: settings to be checked before defaulting to ``os.environ`` or global settings
	:return: None
	'''
	
	# princeps script
	princeps_path = resolve_order(global_settings['princeps_path'], overrides, os.environ)
	if not global_settings['disable_princeps'] and princeps_path is not None:
		try:
			include_files(princeps_path)
		except Exception as e:
			prt.critical(f'Failed to run princeps: {princeps_path}')
			raise e
	
	# load profile
	profile = get_profile(**overrides)
	
	# load project/s
	profile.initialize()
	
	for proj in projects:
		profile.load_project(proj)


def cleanup(**overrides):
	'''
	Cleans up the projects and profile, which by default just updates the project/profile info
	yaml file if new information was added to the project/profile.

	Generally, this should be run after running any desired scripts.

	:param overrides: settings to check before defaulting to global settings or ``os.environ``
	:return: None
	'''
	get_profile(**overrides).cleanup()

# endregion

# region Create

def get_config(*contents, **parameters):
	'''
	Process the provided info using the current project into a config object.
	:param contents: usually a list of parent configs to be merged
	:param parameters: any manual parameters to include in the config object
	:return: config object
	'''
	return get_current_project().create_config(*contents, **parameters)

def create_component(config):
	'''
	Create a component using the current project
	:param config: Must contain a "_type" parameter with the name of a registered component
	:return: the created component
	'''
	return get_current_project().create_component(config)

def quick_create(_type, *parents, **parameters):
	'''
	Creates a component without an explicit config object. Effectively combines `get_config()` and `create_component()`
	:param _type:
	:param parents:
	:param parameters:
	:return:
	'''
	proj = get_current_project()
	
	config = proj.create_config(*parents, **parameters)
	config.push('_type', _type, silent=True)
	
	return proj.create_component(config)
	
# endregion

# region Registration

def register_script(name, fn, description=None, use_config=False):
	'''Manually register a new script to the current project'''
	monkey_patch(fn)
	return get_current_project().register_script(name, fn, description=description, use_config=use_config)

def register_component(name, fn, description=None):
	'''Manually register a new component to the current project'''
	return get_current_project().register_component(name, fn, description=description)

def register_modifier(name, fn, description=None, expects_config=False):
	'''Manually register a new modifier to the current project'''
	return get_current_project().register_modifier(name, fn, description=description, expects_config=expects_config)

def register_config(name, path):
	'''Manually register a new config file to the current project'''
	return get_current_project().register_config(name, path)

def register_config_dir(path, recursive=True, prefix=None, joiner='/'):
	'''Manually register a new config directory to the current project'''
	return get_current_project().register_config_dir(path, recursive=recursive, prefix=prefix, joiner=joiner)

# endregion

# region Artifacts

def has_script(name):
	return get_current_project().has_script(name)
def find_script(name):
	return get_current_project().find_script(name)
def view_scripts():
	return get_current_project().view_scripts()

def has_component(name):
	return get_current_project().has_component(name)
def find_component(name):
	return get_current_project().find_component(name)
def view_components():
	return get_current_project().view_components()

def has_modifier(name):
	return get_current_project().has_modifier(name)
def find_modifier(name):
	return get_current_project().find_modifier(name)
def view_modifiers():
	return get_current_project().view_modifiers()

def has_config(name):
	return get_current_project().has_config(name)
def find_config(name):
	return get_current_project().find_config(name)
def view_configs():
	return get_current_project().view_configs()



# endregion





import sys, os
from pathlib import Path
from copy import deepcopy, copy
import omnibelt as belt
from collections import defaultdict, OrderedDict

from omnibelt import save_yaml, load_yaml, get_printer

from .util import primitives, global_settings, configurize, pythonize, ConfigurizeFailed
from .errors import PythonizeError, MissingParameterError, UnknownActionError, InvalidKeyError

prt = get_printer(__name__)

_printing_instance = None
class Config_Printing:
	'''
	Internal class to manage the printing pulls/pushes of the config object. (eg. indent/line styles)
	'''
	def __new__(cls, *args, **kwargs): # singleton
		global _printing_instance
		if _printing_instance is None:
			_printing_instance = super().__new__(cls)
			
			_printing_instance.level = 0
			_printing_instance.is_new_line = True
			_printing_instance.unit = ' > '
			_printing_instance.style = '| '
			_printing_instance.silent = False
			_printing_instance.src = None
		
		return _printing_instance
	def __repr__(self):
		return f'ConfigPrinting[{hex(id(self))}]'
	def __str__(self):
		return f'ConfigPrinting'
	
	# def set_src(self, src):
	# 	self.src = src
	
	def inc_indent(self):
		self.level += 1
	def dec_indent(self):
		self.level = max(0, self.level-1)
	
	def process_addr(self, *terms):

		addr = []
		
		skip = False
		for term in terms[::-1]:
			if term == '':
				skip = True
				addr.append('')
			elif term is None:
				pass
			elif not skip:
				addr.append(term)
			else:
				skip = False
		
		if not len(addr):
			return '.'
		
		addr = '.'.join(addr[::-1])
		return addr
	
	def log_record(self, raw, end='\n',
	               silent=False):
		indent = self.level * self.unit
		style = self.style
		src = '' if self.src is None else f'({self.src}) '
		prefix = style + src + indent
		
		msg = raw.replace('\n', '\n' + prefix)
		if not self.is_new_line:
			prefix = ''
		msg = f'{prefix}{msg}{end}'
		base = self.silent
		if not (silent or base):
			print(msg, end='')
			self.is_new_line = (len(msg) == 0 and self.is_new_line) or msg[-1] == '\n'
		return msg
		

class ConfigType(belt.Transactionable):
	'''
	The abstract super class of config objects.
	
	The most important methods:
	
		- ``push()`` - set a parameter in the config
		- ``pull()`` - get a parameter in the config
		- ``sub()`` - get a sub branch of this config
		- ``seq()`` - iterate through all contents
		- ``update()`` - update a config with a different config
		- ``export()`` - save the config object as a yaml file
	
	Another important property of the config object is that it acts as a tree where each node can hold parameters.
	If a parameter cannot be found at one node, it will search up the tree for the parameter.
	
	Config objects also allow for "deep" gets/sets, which means you can get and set parameters not just in
	the current node, but any number of nodes deeper in the tree by passing a list/tuple of keys
	or keys separated by "." (hereafter called the "address" of the parameter).
	
	Note: that all parameters must not contain "." and should generally be valid python identifiers
	(strings with no white space that don't start with a number).
	
	'''

	def __init__(self, parent=None, printer=None,
	             prefix=None, safe_mode=False, project=None,
	             data=None):
		'''
		Generally it should not be necessary to create a ConfigType manually, instead use ``get_config()``.
		
		:param parent: parent config used for defaults
		:param printer: printer object to handle printing messages
		:param prefix: initial prefix used for printing
		:param project: project this config responds to
		:param safe_mode: don't save created component instances, unless during a transaction
		:param data: raw parameters to immediately add to this config
		'''
		
		if printer is None:
			printer = Config_Printing()
		self.__dict__['_printer'] = printer
		
		if prefix is None:
			prefix = []
		self.__dict__['_prefix'] = prefix
		self.__dict__['_hidden_prefix'] = prefix.copy()
		self.__dict__['_safe_mode'] = safe_mode
		
		self.__dict__['_project'] = None
		
		self.set_parent(parent)
		# self._set_silent(silent)

		if project is not None:
			self.set_project(project)

		super().__init__()
		
		if data is not None:
			self.update(data)

	def __deepcopy__(self, memodict={}):
		raise NotImplementedError

	def __copy__(self):
		raise NotImplementedError

	def copy(self):
		'''shallow copy of the config object'''
		return copy(self)

	def pythonize(self):
		return pythonize(self)

	@classmethod
	def convert(cls, data, recurse):
		'''used by configurize to turn a nested python object into a config object'''
		return cls(data=[recurse(x) for x in data])
		
	def sub(self, item):
		'''
		Used to get a subbranch of the overall config
		:param item: address of the branch to return
		:return: config object at the address
		'''
		
		# TODO: replace with a pull(raw=True)
		
		val = self.get_nodefault(item)
		
		if isinstance(item, (list, tuple)):
			item = '.'.join(item)
		
		if isinstance(val, ConfigType):
			val._set_prefix(self.get_prefix() + [item])
			val._store_prefix()
		
		return val

	def update(self, other):
		'''
		Used to merge two config nodes (and their children) together
		
		This method must be implemented by child classes depending on how the contents of the node is stored
		
		:param other: config node to overwrite ``self`` with
		:return: None
		'''
		try:
			return super().update(other)
		except AttributeError:
			raise NotImplementedError

	def _record_action(self, action, suffix=None, val=None, silent=False, obj=None,
	                   defaulted=False, pushed=False, _entry=False, _raw=False):
		'''
		Internal function to manage printing out messages after various actions have been taken with this config.
		
		:param action: name of the action (see code for examples)
		:param suffix: suffix of the address (aka. last item in the address)
		:param val: contents at that address
		:param silent: suppress printing a message for this action
		:param obj: object created or used in this action (eg. a newly created component)
		:param defaulted: is a default value being used
		:param pushed: has this value just been pushed
		:param _entry: internal flag used for printing messages
		:return: formatted message corresponding to this action
		'''
		printer = self._get_printer()

		if action == 'defaulted':
			return ''
		
		name = printer.process_addr(*self.get_prefix(), suffix)
		
		if pushed:
			name = '[Pushed] ' + name

		if 'alias' in action:
			return printer.log_record(f'{name} --> ', end='', silent=silent)
		
		origins = ' (by default)' if defaulted else ''
		
		if action == 'removing':
			obj_type = action.split('-')[-1]
			return printer.log_record(f'REMOVING {name}', silent=silent)
		
		if action in {'creating', 'reusing'}:
			
			assert val is not None, 'no info provided'
			
			cmpn_type = val.pull('_type', silent=True)
			mods = val.pull('_mod', None, silent=True)
			
			mod_info = ''
			if mods is not None and len(mods):
				if not isinstance(mods, (list, tuple)):
					mods = mods,
				
				mod_info = ' (mods=[{}])'.format(', '.join(m for m in mods)) if len(mods) > 1 \
					else f' (mod={mods[0]})'
			
			end = ''
			if action == 'reusing':
				assert obj is not None, 'no object provided'
				end = f': id={hex(id(obj))}'
			
			# head = '' if suffix is None else f' {name}'
			head = f' {name}'
			out = printer.log_record(f'{action.upper()}{head} (type={cmpn_type}){mod_info}{origins}{end}',
			                       silent=silent)
			if action == 'creating':
				printer.inc_indent()
			return out
		
		if action in {'iter-dict', 'iter-list'}:
			obj_type = action.split('-')[-1]
			
			head = '' if suffix is None else f'{name} [{obj_type} with {len(val)} item/s]'
			return printer.log_record(f'ITERATOR {head}{origins}', silent=silent)
		
		if action in {'pull-dict', 'pull-list'}:
			
			assert val is not None, 'no obj provided'
			
			obj_type = action.split('-')[-1]
			
			out = printer.log_record(f'{name} [{obj_type} with {len(val)} item/s]', silent=silent)
			printer.inc_indent()
			return out
		
		if action in {'created', 'pulled-dict', 'pulled-list'}:
			assert obj is not None, 'no object provided'
			printer.dec_indent()
			return ''
			# return printer.log_record(f'=> id={hex(id(obj))}', silent=silent)

		pval = None if val is None else (f'[{type(val)}]' if _raw else repr(val))
		
		if action == 'entry': # when pulling dict/list
			assert suffix is not None, 'no suffix provided'
			return printer.log_record(f'({suffix}): ', end='',
			                          silent=(silent or self._get_silent()))
		
		if action == 'pulled':
			head = '' if suffix is None else f'{name}: '
			if _entry:
				head = ''
			return printer.log_record(f'{head}{pval}{origins}', silent=silent)
		
		raise UnknownActionError(action)
		
	def pull(self, item, *defaults, silent=False, ref=False, no_parent=False, as_iter=False, raw=False):
		'''
		Top-level function to get parameters from the config object (including automatically creating components)

		:param item: address of the parameter to get
		:param defaults: default values to use if ``item`` is not found
		:param silent: suppress printing message that this parameter was pulled
		:param ref: if the parameter is a component that has already been created, get a reference to the created component instead of creating a new instance
		:param no_parent: don't default to a parent node if the ``item`` is not found here
		:param as_iter: return an iterator over the selected value (only works if the value is a dict/list)
		:return: processed value of the parameter (or default if ``item`` is not found, or raises a ``MissingConfigError`` if not found)
		'''
		self._reset_prefix()
		return self._pull(item, *defaults, silent=silent, ref=ref, no_parent=no_parent, as_iter=as_iter,
		                  _origin=self, _raw=raw)

	def pull_self(self, name=None, silent=False, as_iter=False, raw=False, ref=False):
		'''
		Process self as a value being pulled.
		
		:param name: Name given to self for printed message
		:param silent: suppress printing message
		:param as_iter: Return self as an iterator (has same effect as calling ``seq()``)
		:return: the processed value of self
		'''
		self._reset_prefix()
		return self._process_val(name, self, silent=silent, reuse=ref, is_self=True, as_iter=as_iter,
		                         _origin=self, _raw=raw)

	def _pull(self, item, *defaults, silent=False, ref=False, no_parent=False, as_iter=False,
	          _defaulted=False, _origin=None, _raw=False, _allow_cousin=True):
		'''
		Internal pull method, should generally not be called manually (unless you know what you're doing)
		
		:param item: remaining address to find
		:param defaults: any default values that can be used if address is not found
		:param silent: suppress messages
		:param ref: return an instance of the value instead of creating a new instance, if one exists
		:param no_parent: do not check for the parameter in the parent
		:param as_iter: return the value as an iterator (only for dicts/lists)
		:param _defaulted: flag that this value was once a provided default (used for printing)
		:param _origin: reference to the original config node that was pulled (some pulls require returing to origin)
		:param _raw: return unprocessed value
		:return: processed value found at ``item`` or a processed default value
		'''
		
		# TODO: change defaults to be a keyword argument providing *1* default, and have item*s* instead,
		#  which are the keys to be checked in order

		if '.' in item:
			item = item.split('.')

		line = []
		if isinstance(item, (list, tuple)):
			line = item[1:]
			item = item[0]

		defaulted = item not in self
		byparent = not self.contains_nodefault(item)
		if no_parent and byparent:
			defaulted = True
		if defaulted:
			try:
				if '__origin_key' in self and _allow_cousin: # cousins
					origin = self['__origin_key']
					if origin is not None:
						parent = self.get_parent()
						if parent is not None:
							grandparent = parent.get_parent()
							if grandparent is not None:
								return grandparent._pull((origin, item), silent=silent, _defaulted=defaulted or _defaulted,
							                        as_iter=as_iter, ref=ref, _origin=_origin, _raw=_raw, _allow_cousin=False)
			except MissingParameterError:
				pass
			
			if len(defaults) == 0:
				raise MissingParameterError(item)
			val, *defaults = defaults
			line = []
		else:
			val = self[item]

		if len(line) and not isinstance(val, ConfigType):
			defaulted = True
			if len(defaults) == 0:
				raise MissingParameterError(item)
			val, *defaults = defaults

		if defaulted and _origin is not None: # try again with new value

			_origin._reset_prefix()
			
			val = _origin._process_val(item, val, *defaults, silent=silent, defaulted=defaulted or _defaulted,
			                        as_iter=as_iter, reuse=ref, _origin=_origin, _raw=_raw)

		elif len(line): # child pull
			if not isinstance(val, ConfigType):
				prt.warning(f'Pulling through a non-config object: {val}')
			
			val._set_prefix(self.get_prefix() + [item])
			out = val._pull(line, *defaults, silent=silent, ref=ref, no_parent=no_parent, as_iter=as_iter,
			               _defaulted=_defaulted, _origin=_origin, _raw=_raw, _allow_cousin=_allow_cousin)
			
			val = out
			
		elif byparent and not item.startswith('_'): # parent pull
			parent = self.get_parent()

			parent._set_prefix(self.get_prefix() + [''])
			val = parent._pull((item, *line), *defaults, silent=silent, ref=ref, no_parent=no_parent, as_iter=as_iter,
			                   _origin=_origin, _raw=_raw, _allow_cousin=_allow_cousin)
			
		else: # process val from me/defaults
			val = self._process_val(item, val, *defaults, silent=silent, defaulted=defaulted or _defaulted,
			                        as_iter=as_iter, reuse=ref, _origin=_origin, _raw=_raw)
			
			if type(val) in {list, set}: # TODO: a little heavy handed
				val = tuple(val)

		return val

	def _process_val(self, item, val, *defaults, silent=False,
	                 reuse=False, is_self=False, as_iter=False, _origin=None, _raw=False,
	                 **record_flags):
		'''
		This is used by ``pull()`` to process the recovered value and print the correct message if ``not silent``
		
		:param item: remaining address where this value was found
		:param val: value that was found given the original address
		:param defaults: any additional defaults to use if processing fails with ``val``
		:param silent: suppress messages
		:param reuse: if an instance is found (under ``__obj``) then that should be returned instead of creating a new one
		:param is_self: this config object should be returned after processing
		:param as_iter: return an iterator of ``val`` (only works if ``val`` is a list/dict)
		:param _origin: original config node where the pull or push request was intially called
		:param record_flags: additional flags used for printing.
		:return: processed value
		'''
		
		if as_iter and isinstance(val, ConfigType):

			obj_type = 'list' if isinstance(val, list) else 'dict'

			if obj_type == 'list' or '_type' not in val or not val.pull('_type', silent=True).startswith('iter'):

				self._record_action(f'iter-{obj_type}', suffix=item, val=val, silent=silent, **record_flags)

				if _raw:
					return val

				itr = ConfigIter(val, val)

				return itr
		
		if isinstance(val, ConfigDict) and not _raw:
			
			val.push('__origin_key', item, silent=True)
			typ = val._pull('_type', None, silent=True)
			if typ is not None:
				
				if reuse and '__obj' in val:
					# print('WARNING: would usually reuse {} now, but instead creating a new one!!')
					cmpn = val['__obj']
					self._record_action('reusing', suffix=item, val=val, obj=cmpn, silent=silent, **record_flags)
				else:

					self._record_action('creating', suffix=item, val=val, silent=silent, **record_flags)
					
					hidden = val._get_hidden_prefix()
					val.get_prefix().clear()
					val._store_prefix()
					
					past = self._get_silent()
					with self.silenced(silent or past):
						cmpn = self.get_project().create_component(val)

					if not self.in_safe_mode() or self.in_transaction():
						if item is not None and len(item) and not is_self:
							val['__obj'] = cmpn
						else:
							self['__obj'] = cmpn
					
					val._set_prefix(hidden)
					val._store_prefix()
					
					self._record_action('created', suffix=item, val=val, obj=cmpn, silent=silent, **record_flags)
					
				val = cmpn

			else:
				
				val.push('__origin_key', '_x_', silent=True)
				
				self._record_action('pull-dict', suffix=item, val=val, silent=silent, **record_flags)
				terms = {}
				for k, v in val.items():  # WARNING: pulls all entries in dict
					k = str(k)
					self._record_action('entry', silent=silent, suffix=k)
					terms[k] = val._process_val(k, v, reuse=reuse, silent=silent, _origin=_origin, _entry=True)
				self._record_action('pulled-dict', suffix=item, val=val, obj=terms, silent=silent, **record_flags)
				val = terms


		elif isinstance(val, ConfigList) and not _raw:
			
			self._record_action('pull-list', suffix=item, val=val, silent=silent, **record_flags)
			terms = []
			for i, v in enumerate(val):  # WARNING: pulls all entries in list
				self._record_action('entry', silent=silent, suffix=str(i))
				terms.append(val._process_val(str(i), v, reuse=reuse, silent=silent, _origin=_origin, _entry=True))
			self._record_action('pulled-list', suffix=item, val=val, obj=terms, silent=silent, **record_flags)
			val = terms

		elif isinstance(val, str) and val.startswith('<>'):  # local alias (looks for alias locally)
			alias = val[2:]
			
			self._record_action('local-alias', suffix=item, val=alias, silent=silent, **record_flags)
			
			val = self._pull(alias, *defaults, silent=silent, _origin=_origin, as_iter=as_iter, ref=reuse)
			
		elif isinstance(val, str) and val.startswith('<o>'): # origin alias (returns to origin to find alias)
			alias = val[3:]
			
			self._record_action('origin-alias', suffix=item, val=alias, silent=silent, **record_flags)
			
			_origin._reset_prefix()
			val = _origin._pull(alias, *defaults, silent=silent, _origin=_origin, as_iter=as_iter)

		elif isinstance(val, str) and val.startswith('<!>'): # copy alias (only for local aliases)
			alias = val[3:]

			self._record_action('copy-alias', suffix=item, val=alias, silent=silent, **record_flags)

			val = self._pull(alias, *defaults, silent=silent, _origin=_origin, as_iter=as_iter, _raw=True)

			# self[item] = deepcopy(val)
			val = deepcopy(val)
			self[item] = val

			return self._process_val(item, val, silent=silent, _origin=_origin, as_iter=as_iter, _raw=_raw)


		else:
			self._record_action('pulled', suffix=item, val=val, silent=silent, _raw=_raw, **record_flags)

		return val
	
	def push(self, key, val, *_skip, silent=False, overwrite=True, no_parent=True, force_root=False, process=True):
		'''
		Set ``key`` with ``val`` in the config object, but pulls ``key`` first so that `val` is only set
		if it is not found or ``overwrite`` is set to ``True``. It will return the current value of ``key`` after
		possibly setting with ``val``.

		:param key: key to check/set (can be list or '.' separated string)
		:param val: data to possibly write into the config object
		:param _skip: soak up all other positional arguments to make sure the remaining are keyword only
		:param silent: Do not print messages
		:param overwrite: If key is already set, overwrite with (configurized) 'val'
		:param no_parent: Do not check parent object if not found in self
		:param force_root: Push key to the root config object
		:return: current val of key (updated if written)
		'''
		self._reset_prefix()
		return self._push(key, val, silent=silent, overwrite=overwrite, no_parent=no_parent,
		                  force_root=force_root, process=process)
	
	def _push(self, key, val, silent=False, overwrite=True, no_parent=True, force_root=False,
	          process=True, _origin=None):
		'''
		Set ``key`` with ``val`` in the config object, but pulls ``key`` first so that `val` is only set
		if it is not found or ``overwrite`` is set to ``True``. It will return the current value of ``key`` after
		possibly setting with ``val``.
		
		:param key: key to check/set (can be list or '.' separated string)
		:param val: data to possibly write into the config object
		:param _skip: soak up all other positional arguments to make sure the remaining are keyword only
		:param silent: Do not print messages
		:param overwrite: If key is already set, overwrite with (configurized) 'val'
		:param no_parent: Do not check parent object if not found in self
		:param force_root: Push key to the root config object
		:return: current val of key (updated if written)
		'''
		
		if '.' in key:
			key = key.split('.')

		line = []
		if isinstance(key, (list, tuple)):
			line = key[1:]
			key = key[0]
		
		exists = self.contains_nodefault(key)
		
		parent = self.get_parent()
		
		if parent is not None and force_root:
			return self.get_root().push((key, *line), val, silent=silent, overwrite=overwrite,
			                   no_parent=no_parent, process=process)
		elif no_parent:
			parent = None
		
		if exists and len(line): # push me
			child = self.get_nodefault(key)
			
			child._set_prefix(self.get_prefix() + [key])
			
			out = child._push(line, val, silent=silent, overwrite=overwrite, no_parent=True, process=process)
			
			return out
		elif parent is not None and key in parent: # push parent

			parent._set_prefix(self.get_prefix() + [''])
			
			out = parent._push((key, *line), val, silent=silent, overwrite=overwrite,
			                   no_parent=no_parent, process=process)
			
			return out
			
		elif len(line): # push child
			
			child = self.get_nodefault(key)
			
			child._set_prefix(self.get_prefix() + [key])
			
			out = child._push(line, val, silent=silent, overwrite=overwrite, no_parent=True, process=process)
		
			return out
		
		if exists and not overwrite:
			return self._pull(key, silent=True)
		
		if isinstance(val, str) and val == '_x_':
			if exists:
				self._record_action('removing', suffix=key, val=val, silent=silent)
				del self[key]
			return
		
		val = configurize(val)
		
		self[key] = val
		# val = self[key]
		
		if process:
			val = self._process_val(key, val, silent=silent, pushed=True)
		
		return val

	def export(self, path=None):
		'''
		Convert all data to raw data (using dict/list) and save as yaml file to ``path`` if provided.
		Also returns yamlified data.
		
		:param path: path to save data in this config (data is not saved to disk if not provided)
		:return: raw "yamlified" data
		'''

		data = pythonize(self)

		if path is not None:
			path = Path(path)
			if path.is_dir():
				path = path / 'config.yaml'
			save_yaml(data, path)
			return path

		return data
	
	def seq(self):
		'''
		Returns an iterator over the contents of this config object where elements are lazily
		processed during iteration (see ``ConfigIter`` for details).
		
		:return: iterator over all arguments in self
		'''
		return ConfigIter(self, self)

	def replace_vals(self, replacements):
		raise NotImplementedError

	# region Silencing
	
	def set_silent(self, silent=True):
		'''Sets whether pushes and pulls on this config object should be printed out to stdout'''
		self.__dict__['_printer'].silent = silent
	
	def silence(self, silent=True):
		self.set_silent(silent)
	
	# self._silent_config_flag = silent
	
	def _get_silent(self):
		return self.__dict__['_printer'].silent
	
	class _Silent_Config:
		'''Internal context manager to silence a config object'''
		
		def __init__(self, config, setting):
			self.config = config
			self.setting = setting
			self.prev = config._get_silent()
		
		def __enter__(self):
			self.config.set_silent(self.setting)
			return self.config
		
		def __exit__(self, exc_type, exc_val, exc_tb):
			self.config.set_silent(self.prev)
	
	def silenced(self, setting=True):
		'''Returns a context manager to silence this config object'''
		return ConfigType._Silent_Config(self, setting=setting)
	
	# endregion
	
	# region Parents
	
	def is_root(self):  # TODO: move to tree space
		'''Check if this config object has a parent for defaults'''
		return self.get_parent() is None
	
	def set_parent(self, parent):
		'''Sets the parent config object to be checked when a parameter is not found in `self`'''
		self.__dict__['_parent'] = parent
	
	def get_parent(self):
		'''Get parent (returns None if this is the root)'''
		return self.__dict__['_parent']
	
	def set_process_id(self, name=None):
		'''Set the unique ID to include when printing out pulls from this object'''
		self._get_printer().src = name
	
	def get_root(self):
		'''Gets the root config object (returns ``self`` if ``self`` is the root)'''
		parent = self.get_parent()
		if parent is None:
			return self
		return parent.get_root()
	
	# endregion
	
	# region Addressing
	
	def _get_printer(self):
		return self.__dict__['_printer']
	
	def get_prefix(self):
		return self.__dict__['_prefix']
	
	def _set_prefix(self, prefix):
		self.__dict__['_prefix'] = prefix
	
	def _reset_prefix(self):
		self._set_prefix(self._get_hidden_prefix())
	
	def _get_hidden_prefix(self):
		return self.__dict__['_hidden_prefix']
	
	def _store_prefix(self):
		self.__dict__['_hidden_prefix'] = self.get_prefix().copy()
	
	# endregion
	
	# region Misc
	
	def set_project(self, project):
		self.get_root().__dict__['_project'] = project
	
	def get_project(self):
		return self.get_root().__dict__['_project']
	
	def set_safe_mode(self, safe_mode):
		if self.is_root():
			self.__dict__['_safe_mode'] = safe_mode
		else:
			self.get_root().set_safe_mode(safe_mode)
	
	def in_safe_mode(self):
		if self.is_root():
			return self.__dict__['_safe_mode']
		return self.get_root().in_safe_mode()
	
	def purge_volatile(self):
		'''
		Recursively remove any items where the key starts with "__"

		Must be implemented by the child class

		:return: None
		'''
		raise NotImplementedError
	
	def __repr__(self):
		return f'{type(self).__name__}[id={hex(id(self))}]'
	
	def __str__(self):
		# return repr(self)
		return f'<{type(self).__name__}>'
	
	# endregion
	
	# region Get/Set/Contains
	
	def __setitem__(self, key, value):
		if isinstance(key, str) and '.' in key:
			key = key.split('.')
		
		if isinstance(key, (list, tuple)):
			if len(key) == 1:
				return self.__setitem__(key[0], value)
			child = self.get_nodefault(*key)
			assert isinstance(child, ConfigType)
			return child.__setitem__(key[1:], value)
		
		value = configurize(value)
		
		if isinstance(value, ConfigType):
			value.set_parent(self)
			# value.set_project(self.get_project())
		return self._single_set(key, value)
	
	def __getitem__(self, item, *future):
		
		if isinstance(item, str) and '.' in item:
			item = item.split('.')
		
		if isinstance(item, (list, tuple)):
			if len(item) == 1:
				item = item[0]
			else:
				return self.__getitem__(*item)[item[1:]]
		
		parent = self.get_parent()
		
		if not self.contains_nodefault(item) \
				and parent is not None \
				and item[0] != '_':
			return parent[item]
		
		return self._single_get(item, *future)
	
	def get_nodefault(self, item, *future):
		'''Get ``item`` without defaulting up the tree if not found.'''
		
		if isinstance(item, str) and '.' in item:
			item = item.split('.')
		
		if isinstance(item, (list, tuple)):
			if len(item) == 1:
				item = item[0]
			else:
				return self.get_nodefault(*item)[item[1:]]
		
		return self._single_get(item, *future)
	
	def _single_set(self, key, val):
		return super().__setitem__(key, val)
	
	def _single_get(self, item, *context):
		try:
			val = super().__getitem__(item)
			if val is EmptyElement and len(context):
				raise KeyError(item)
		except KeyError:
			return self._missing_key(item, *context)
		
		return val
	
	def _missing_key(self, key, *context):
		
		cls = ConfigDict
		
		if len(context):
			nxt = context[0]
			try:
				int(nxt)
			except ValueError:
				pass
			else:
				cls = ConfigList
		
		obj = cls(parent=self)
		self.__setitem__(key, obj)
		return obj
	
	def __contains__(self, item):
		'''Check if ``item`` is in this config, item can be "deep" (multiple steps'''
		if isinstance(item, str) and '.' in item:
			item = item.split('.')
		
		if isinstance(item, (tuple, list)):
			if len(item) == 1:
				item = item[0]
			else:
				return item[0] in self and item[1:] in self[item[0]]
		
		parent = self.get_parent()
		
		return self.contains_nodefault(item) \
		       or (not super().__contains__(item)
		           and parent is not None
		           and item[0] != '_'
		           and item in parent)
	
	def contains_nodefault(self, item):
		'''Check if ``item`` is contained in this config object without defaulting up the tree if ``item`` is not found'''
		
		if isinstance(item, str) and '.' in item:
			item = item.split('.')
		
		if isinstance(item, (tuple, list)):
			if len(item) == 1:
				item = item[0]
			else:
				return self.contains_nodefault(item[0]) and self[item[0]].contains_nodefault(item[1:])
		
		if super().__contains__(item):
			return self.get_nodefault(item) is not '__x__'
		return False

	# endregion


class ConfigDict(ConfigType, belt.tdict):
	'''
	Dict like node in the config.
	
	Keys should all be valid python attributes (strings with no whitespace, and not starting with a number).

	NOTE: avoid setting keys that start with more than one underscore (especially '__obj')
	(unless you really know what you are doing)
	'''

	def __deepcopy__(self, memodict={}):
		new = self.__class__(data={k:deepcopy(v) for k,v in self.items() if not k.startswith('__')})
		new.__dict__.update(self.__dict__)
		return new

	def __copy__(self):
		new = self.__class__(data={k:v for k,v in self.items()})
		new.__dict__.update(self.__dict__)
		return new

	def copy(self):
		return copy(self)

	def replace_vals(self, replacements):
		for k,v in self.items():
			if isinstance(v, primitives) and v in replacements:
				self[k] = replacements[v]
			elif isinstance(v, ConfigType):
				v.replace_vals(replacements)

	@classmethod
	def convert(cls, data, recurse):
		return cls(data={k: recurse(v) for k, v in data.items()})
	
	def update(self, other):
		'''
		Merge self with another dict-like object
		
		:param other: must be dict like
		:return: None
		'''
		assert isinstance(other, dict), f'invalid: {type(other)}'
		
		for k, v in other.items():
			isconfig = isinstance(v, ConfigType)
			exists = self.contains_nodefault(k)
			if exists and '_x_' == v:  # reserved for deleting settings in parents
				del self[k]
			elif exists and isconfig and \
					(isinstance(v, self[k].__class__)
					 or isinstance(self[k], v.__class__)):
				self[k].update(v)
				
			else:
				self[k] = v
	
	def purge_volatile(self):
		'''
		Recursively remove any items where the key starts with "__"
		
		:return: None
		'''
		bad = []
		for k, v in self.items():
			if k.startswith('__'):
				bad.append(k)
			elif isinstance(v, ConfigType):
				v.purge_volatile()
		
		for k in bad:
			del self[k]
	
	def __repr__(self):
		info = '{{' + ', '.join(f'{k}' for k in self) + '}}'
		return f'[{hex(id(self))}]{info}'
	
	def __str__(self):
		return repr(self)
		return '{{' + ', '.join(f'{k}' for k in self) + '}}'


class EmptyElement:
	pass

class ConfigList(ConfigType, belt.tlist):
	'''
	List like node in the config.
	'''



	def __init__(self, *args, empty_fill_value=EmptyElement, **kwargs):
		super().__init__(*args, **kwargs)
		self._empty_fill_value = empty_fill_value

	def __deepcopy__(self, memodict={}):
		new = self.__class__(data=[deepcopy(x) for x in self])
		new.__dict__.update(self.__dict__)
		return new

	def __copy__(self):
		new = self.__class__(data=[x for x in self])
		new.__dict__.update(self.__dict__)
		return new

	def __repr__(self):
		info = '[[' + ', '.join(f'{k}' for k in self) + ']]'
		return f'[{hex(id(self))}]{info}'

	def __str__(self):
		return '[[' + ', '.join(f'{k}' for k in self) + ']]'

	def replace_vals(self, replacements):
		for i, x in enumerate(self):
			if isinstance(x, primitives) and x in replacements:
				self[i] = replacements[x]
			elif isinstance(x, ConfigType):
				x.replace_vals(replacements)


	def purge_volatile(self):
		'''
		Recursively remove any items where the key starts with "__"
		
		:return: None
		'''
		for x in self:
			if isinstance(x, ConfigType):
				x.purge_volatile()
		
	def _str_to_int(self, item):
		'''Convert the input items to indices of the list'''
		if isinstance(item, int):
			return item

		try:
			return int(item)
		except (TypeError, ValueError):
			pass

		raise InvalidKeyError(f'failed to convert {item} to an index')
		
	def update(self, other):
		'''Overwrite ``self`` with the provided list ``other``'''
		for i, x in enumerate(other):
			isconfig = isinstance(x, ConfigType)
			if len(self) == i:
				self.append(x)
			elif isconfig and (isinstance(x, self[i].__class__) or isinstance(self[i], x.__class__)):
				self[i].update(x)
			elif x is not EmptyElement:
				self[i] = x
	

	def _single_set(self, key, val):
		if key == '_':  # append to end of the list
			key = len(self)
		key = self._str_to_int(key)
		if key >= len(self):
			self.extend([self._empty_fill_value] * (key - len(self) + 1))
		return super()._single_set(key, val)
	
	def _single_get(self, item, *context):
		
		if isinstance(item, slice):
			return super(ConfigType, self).__getitem__(item)
		
		if item == '_':  # append to end of the list
			item = len(self)
		
		item = self._str_to_int(item)
		
		if item < 0:
			item += len(self)
		if item >= len(self):
			self.extend([self._empty_fill_value] * (item - len(self) + 1))
			
		return super()._single_get(item, *context)

	def push(self, first, *rest, overwrite=True, **kwargs):
		'''
		When pushing to a list, if you don't provide an index, the value is automatically pushed to the end of the list
		
		:param first: if no additional args are provided in `rest`, then this is used as the value and the key is the end of the list, otherwise this is used as key and the first element in `rest` is the value
		:param rest: optional second argument to specify the key, rather than defaulting to the end of the list
		:param kwargs: same keyword args as for the ConfigDict
		:return: same as for ConfigDict
		'''

		if len(rest):
			val, *rest = rest
			return super().push(first, val, *rest, overwrite=overwrite, **kwargs)
		
		val = first
		key = len(self)
		self.append(None)
		return super().push(key, val, *rest, overwrite=True, **kwargs) # note that *rest will have no effect
		
	def contains_nodefault(self, item):
		
		try:
			idx = self._str_to_int(item)
		except InvalidKeyError:
			return isinstance(item, slice)
		
		N = len(self)
		return -N <= idx < N
		
	def append(self, item):
		super().append(item)
		if isinstance(item, ConfigType):
			item.set_parent(self)
			# item.set_project(self.get_project())
		
	def extend(self, item):
		super().extend(item)
		
		for x in item:
			if isinstance(x, ConfigType):
				x.set_parent(self)
				# x.set_project(self.get_project())


class ConfigIter:
	'''
	Iterate through a list of parameters, processing each item lazily,
	ie. only when it is iterated over (with ``next()``)
	'''
	
	def __init__(self, origin, elements=None, auto_pull=True, include_key=None, reversed=False):
		'''
		Can be used as a component or created manually (by providing the ``elements`` argument explicitly)

		For dicts, this will behave like ``.items()``, ie. for each entry in the dict it will return
		a tuple of the key and value.

		:param origin: config object where the iterator info is
		:param elements: manually provided elements to iterate over (uses contents of "_elements" in ``origin`` if not provided)
		'''
		# self._name = config._ident
		# assert '_elements' in config, 'No elements found'
		if elements is None:
			elements = origin['_elements']
		self._elms = elements
		
		self._keys = [k for k in self._elms.keys()
		              if k not in {'_elements', '_mod', '_type', '__obj', '__origin_key'}
		              and self._elms[k] != '__x__'] \
			if isinstance(self._elms, dict) else None
		self._include_key = include_key if include_key is not None else self._keys is not None
		self._prefix = origin.get_prefix().copy()
		
		self._reversed = False
		self._idx = 0
		self.set_reversed(reversed)
		self.set_auto_pull(auto_pull)
	
	def __len__(self):
		'''Returns the remaining length of this iterator instance'''
		return len(self._elms if self._keys is None else self._keys) - self._idx
	
	def _next_idx(self):
		'''Find the next index or key'''
		
		if self._keys is None:
			if 0 > self._idx >= len(self._elms):
				raise StopIteration
			return str(self._idx)
			# if not self._elms.contains_nodefault(self._idx):
			# 	raise StopIteration
			# return str(self._idx)
		
		while 0 <= self._idx < len(self._elms):
			idx = self._keys[self._idx]
			
			if self._elms.contains_nodefault(idx):
				return idx
			self._idx += (-1)**self._reversed
		
		raise StopIteration
	
	def view(self):
		'''Returns the next object without processing the item, may throw a StopIteration exception'''
		idx = self._next_idx()
		obj = self._elms.pull(idx, raw=True, silent=True)
		# obj = self._elms[idx]
		if isinstance(obj, ConfigType):
			obj.push('_iter_key', idx, silent=True)
			
		if isinstance(obj, global_settings['config_type']):
			obj = self._elms.sub(idx)
		return (idx, obj) if self._include_key else obj

	def step(self):
		obj = self.view()
		self._idx += (-1)**self._reversed
		return obj

	def set_auto_pull(self, auto=True):
		self._auto_pull = auto

	def set_reversed(self, reversed=True):
		self._reversed = reversed
		if reversed:
			self._idx = len(self)-1

	def has_next(self):
		return (self._reversed and self._idx >= 0) or (not self._reversed and self._idx < len(self._elms))
 
	def __next__(self):
		
		if not self.has_next():
			raise StopIteration

		obj = self.step()
		key, val = obj if self._include_key else (None,obj)
		if isinstance(val, global_settings['config_type']):
			val = val.pull_self(raw=not self._auto_pull, silent=not self._auto_pull)
		return (key,val) if self._include_key else val
	
	def __iter__(self):
		return self

nones = {'None', 'none', '_none', '_None', 'null', 'nil', }
def configurize_nones(s, recurse):
	'''Turns strings into None, if they match the expected patterns'''
	if s in nones:
		return None
	raise ConfigurizeFailed

global_settings.update({
	'config_type': ConfigType,
	'config_converters': OrderedDict([
		(str, configurize_nones),
		(dict, (False, ConfigDict.convert)),
		(OrderedDict, (False, ConfigDict.convert)),
		(list, ConfigList.convert),
		# (tuple, ConfigList.convert),
		(set, ConfigList.convert),
	]),
})


